#include "core.h"
#include <math.h>
#include "SDL/SDL_opengl.h"
#include "dtime.h"
#include <unistd.h>

SimCore::SimCore()
{
	camera_mode = CAMERA_FOLLOW_ROBOT_TS;
	p2m = 100;

	scaleMult = 1.0;
	
	robot = new Robot;
	objects.push_back(robot);
	ground = new Ground;
	objects.push_back(ground);

	mousex = mousey = 0;
	ac = AudioCore::AC();
}

SimCore::~SimCore()
{
	// all deleted when start finishes
}

void SimCore::SDL_GL_Setup()
{
	SDL_Init(SDL_INIT_EVERYTHING);

	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	//SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);

	screen = SDL_SetVideoMode(sW, sH, 0, SDL_OPENGL);

	SDL_EnableKeyRepeat(500, 30);
	SDL_EnableUNICODE(1);

	SetWindowPriTitle("simple_sim");

	ac->SetupCore(44100, 2);
	ac->LoadSingleWAV("media/mine.wav");

	glClearColor(0.0, 0.0, 0.0, 0.0);

	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

void SimCore::SetWindowPriTitle(string s)
{
	pri_title = s;
	UpdateTitle();
}

void SimCore::SetWindowSecondaryTitle(string s)
{
	sec_title = s;
	UpdateTitle();
}

void SimCore::SetLoggerFile(string lf)
{
	logfile = lf;
}

void SimCore::SetUserParamsArg(string ucp)
{
	ucparams = ucp;
}

void SimCore::UpdateTitle()
{
	char tmp[200];
	
	if(sec_title != "") {
		snprintf(tmp, 200, "%s (%s)", pri_title.c_str(), sec_title.c_str());
	} else {
		snprintf(tmp, 200, "%s", pri_title.c_str());
	}

	SDL_WM_SetCaption(tmp, tmp);
}

void SimCore::GetRobotPosVec(double *v)
{
	v[0] = robot->x;
	v[1] = robot->y;
	v[2] = robot->z;
	v[3] = robot->theta;
}

void SimCore::Start()
{
	SDL_GL_Setup();

	wc = WorldClock::WC();

	ODEif = new ODEInterface(1.0/100.0);

	list<Object *>::iterator it;
	for(it = objects.begin(); it != objects.end(); it++) {
		ODEif->RegisterObject(*it);
		(*it)->SetupGL();
	}

	done = false;

	user_if = new UCInterface(1/10.0, ucparams, logfile, this);
	
	wc->SetPaused(false);

	while(!done) {
		SDL_Event ev;
		while(SDL_PollEvent(&ev)) {
			if(ev.type == SDL_QUIT) {
				done = true;
			} else if(ev.type == SDL_MOUSEMOTION) {
				mousex = ev.motion.x;
				mousey = ev.motion.y;
				WorldMousePosition();
			} else if(ev.type == SDL_MOUSEBUTTONDOWN) {
				if(WorldMousePosition()) {
					int button;
					switch(ev.button.button) {
					case SDL_BUTTON_LEFT: button=1; break;
					case SDL_BUTTON_MIDDLE: button=2; break;
					case SDL_BUTTON_RIGHT: button=3; break;
					case SDL_BUTTON_WHEELDOWN: button=4; break;
					case SDL_BUTTON_WHEELUP: button=5; break;
					}
					user_if->mouse_click((int)mousex, (int)mousey, worldx, worldy, button);
				}
			} else if(ev.type == SDL_KEYDOWN) {
				switch(ev.key.keysym.sym) {
				case SDLK_EQUALS:
					scaleMult -= 0.05;
					break;
				case SDLK_MINUS:
					scaleMult += 0.05;;
					break;
				case SDLK_TAB:
					ToggleCameraMode();
					break;
				case SDLK_LEFT:
					camx -= 0.2;
					break;
				case SDLK_RIGHT:
					camx += 0.2;
					break;
				case SDLK_UP:
					camy += 0.2;
					break;
				case SDLK_DOWN:
					camy -= 0.2;
					break;
				case SDLK_d:
					fprintf(stderr, "dumping world\n");
					ODEif->DumpWorld("world.dump");
					break;
				case SDLK_t:
					robot->ToggleTrail();
					break;
				case SDLK_p:
					if(wc->IsPaused()) {
						fprintf(stderr, "Resuming simulation\n");
						wc->SetPaused(false);
					} else {
						fprintf(stderr, "Pausing simulation\n");
						wc->SetPaused(true);
					}
					break;
				case SDLK_w:
					if(wc->GetTimeMultiplier() >= 0.2) {
						double nm = wc->GetTimeMultiplier() - 0.1;
						fprintf(stderr, "Decreasing simulation speed to %.2lfx\n", nm);
						wc->SetTimeMultiplier(nm);
					}
					break;
				case SDLK_e:
					if(wc->GetTimeMultiplier() < 10.0) {
						double nm = wc->GetTimeMultiplier() + 0.1;
						fprintf(stderr, "Increasing simulation speed to %.2lfx\n", nm);
						wc->SetTimeMultiplier(nm);
					}
					break;
				case SDLK_r:
					fprintf(stderr, "Setting simulation speed to 1.0x\n");
					wc->SetTimeMultiplier(1.0);
					break;
				default:
					//fprintf(stderr, "keys=%i mod=%i\n",
					//		ev.key.keysym.sym,
					//		ev.key.keysym.mod);
					if(ev.key.keysym.unicode != 0 && (ev.key.keysym.unicode & 0xFF80) == 0) {
						char ch = ev.key.keysym.unicode & 0x7F;
						user_if->key_press(ch);
					}
					break;
				}
			}
		}

		if(scaleMult <= 0.0) scaleMult = 0.05;

		list<Object *>::iterator it;
		for(it = objects.begin(); it != objects.end(); it++) {
			Object *o = *it;
			if(o->IsDeleted()) {
				// ODEif has declared it dead, safe to kill
				it = objects.erase(it);
				it--;
				delete o;
			}
		}

		// look for collisions
		int rc;
		while(robot->GetCollision(rc)) {
			user_if->collision(rc);
		}

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		SetupCamera();

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		DrawObjects();
		
		SDL_GL_SwapBuffers();
	}
	
	delete user_if;
	delete ODEif;
	for(it = objects.begin(); it != objects.end(); it++) {
		delete *it;
	}

	ac->Die();

	SDL_Quit();
}

void SimCore::SetupCamera()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double swim = (double)sW / p2m;
	double shim = (double)sH / p2m;
	swim *= scaleMult;
	shim *= scaleMult;
	glOrtho(-swim/2, swim/2, -shim/2, shim/2, -1, 1);

	if(camera_mode == CAMERA_FIXED) {
		glTranslated(-camx, -camy, 0);
	} else if(camera_mode == CAMERA_FOLLOW_ROBOT_TS) {
		glTranslated(-robot->x, -robot->y, 0);
		camx = robot->x;
		camy = robot->y;
	} else if(camera_mode == CAMERA_FOLLOW_ROBOT_TANDR) {
		glRotated(180/M_PI * robot->theta, 0, 0, 1);
		glTranslated(-robot->x, -robot->y, 0);
		camx = robot->x;
		camy = robot->y;
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_X) {
		glLoadIdentity();
		gluPerspective(90, (double)sW / (double)sH, 0.5, 50.0);
		gluLookAt(robot->x, -1, 1, robot->x, robot->y, 0, 0, 0, 1);
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_Y) {
		glLoadIdentity();
		gluPerspective(90, (double)sW / (double)sH, 0.5, 50.0);
		gluLookAt(-1, robot->y, 1, robot->x, robot->y, 0, 0, 0, 1);
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_XY) {
		glLoadIdentity();
		gluPerspective(90, (double)sW / (double)sH, 0.5, 50.0);
		double cx, cy;
		cx = robot->x - cos(-robot->theta)*1;
		cy = robot->y - sin(-robot->theta)*1;
		gluLookAt(cx, cy, 0.75, robot->x, robot->y, 0, 0, 0, 1);
	}
}

bool SimCore::WorldMousePosition()
{
	double sW = this->sW, sH = this->sH;

	if(/*camera_mode == CAMERA_FOLLOW_ROBOT_TANDR ||*/
	   camera_mode == CAMERA_FOLLOW_ROBOT_TS ||
	   camera_mode == CAMERA_FIXED) {
		double shim = sH / p2m * scaleMult;
		double swim = sW / p2m * scaleMult;
		worldx = camx - 0.5*swim + swim * (mousex/sW);
		worldy = camy + 0.5*shim - shim * (mousey/sH);
		if(camera_mode == CAMERA_FOLLOW_ROBOT_TANDR) {
			// hmm.. not sure now
		}

		return true;
	}

	return false;
}

void SimCore::ToggleCameraMode()
{
	if(camera_mode == CAMERA_FIXED) {
		camera_mode = CAMERA_FOLLOW_3D_TRACK_X;
		fprintf(stderr, "Now following robot, 3D view, tracking X position\n");
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_X) {
		camera_mode = CAMERA_FOLLOW_3D_TRACK_Y;
		fprintf(stderr, "Now following robot, 3D view, tracking Y position\n");
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_Y) {
		camera_mode = CAMERA_FOLLOW_3D_TRACK_XY;
		fprintf(stderr, "Now following robot, 3D view, behind robot\n");
	} else if(camera_mode == CAMERA_FOLLOW_3D_TRACK_XY) {
		camera_mode = CAMERA_FOLLOW_ROBOT_TANDR;
		fprintf(stderr, "Now following robot translation and rotation\n");
	} else if(camera_mode == CAMERA_FOLLOW_ROBOT_TANDR) {
		camera_mode = CAMERA_FOLLOW_ROBOT_TS;
		fprintf(stderr, "Now following robot translation\n");
	} else if(camera_mode == CAMERA_FOLLOW_ROBOT_TS) {
		camera_mode = CAMERA_FIXED;
		fprintf(stderr, "Now using fixed camera\n");
		camx = robot->x;
		camy = robot->y;
	}
}

void SimCore::SetWindowSize(int x, int y)
{
	sW = x; sH = y;
}

void SimCore::DrawObjects()
{
	list<Object *>::iterator it;
	for(it = objects.begin(); it != objects.end(); it++) {
		Object *i = *it;
		glPushMatrix();
		i->Draw();
		glPopMatrix();
	}
}

